/*
 * MikroSDK - MikroE Software Development Kit
 * Copyright (c) 2019, MikroElektronika - www.mikroe.com
 * All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*!
 * \file oledc.c
 *
 */

#include <string.h>
#include "oledc.h"


// ------------------------------------------------------------- PRIVATE MACROS

#define OLEDC_DEFAULT_MUX_RATIO    95
#define OLEDC_DEFAULT_START_LINE   0x80
#define OLEDC_DEFAULT_OFFSET       0x20

#define OLEDC_DEFAULT_OLED_LOCK    0x12
#define OLEDC_DEFAULT_CMD_LOCK     0xB1
#define OLEDC_DEFAULT_DIVSET       0xF1
#define OLEDC_DEFAULT_PRECHARGE    0x32
#define OLEDC_DEFAULT_VCOMH        0x05
#define OLEDC_DEFAULT_MASTER_CONT  0xCF
#define OLEDC_DEFAULT_PRECHARGE_2  0x01

// ------------------------------------------------------------------ CONSTANTS

static const uint8_t OLEDC_DEFAULT_REMAP = OLEDC_RMP_INC_HOR |
                                           OLEDC_RMP_COLOR_REV |
                                           OLEDC_RMP_SEQ_RGB |
                                           OLEDC_RMP_SCAN_REV |
                                           OLEDC_RMP_SPLIT_ENABLE |
                                           OLEDC_COLOR_65K;

static const uint8_t OLEDC_DEFAULT_VSL[ 3 ] = { 0xA0, 0xB5, 0x55 };
static const uint8_t OLEDC_DEFAULT_CONTRAST[ 3 ] = { 0x8A, 0x51, 0x8A };

// ------------------------------------------------------------------ VARIABLES

static uint8_t oledc_cols[ 2 ] = { OLEDC_COL_OFF, OLEDC_COL_OFF + 95 };
static uint8_t oledc_rows[ 2 ] = { OLEDC_ROW_OFF, OLEDC_ROW_OFF + 95 };

// ---------------------------------------------- PRIVATE FUNCTION DECLARATIONS

static void
pixel( oledc_t *ctx, uint8_t col, uint8_t row, uint16_t color );

static void
character( oledc_t *ctx, uint16_t ch );

static void
box_area( oledc_t *ctx, uint8_t start_col, uint8_t start_row, uint8_t end_col,
          uint8_t end_row, uint16_t color );

static void
draw_area( oledc_t *ctx, uint8_t start_col, uint8_t start_row, uint8_t end_col,
           uint8_t end_row, const uint8_t *img );

// ------------------------------------------------ PUBLIC FUNCTION DEFINITIONS

void
oledc_enable( oledc_t *ctx, uint8_t state )
{
    digital_out_write( &ctx->en, state );
}

void
oledc_reset( oledc_t *ctx )
{
    digital_out_high( &ctx->rst );
    Delay_1ms( );
    digital_out_low( &ctx->rst );
    Delay_1ms( );
    digital_out_high( &ctx->rst );
    Delay_100ms( );
}

void
oledc_command( oledc_t *ctx, uint8_t cmd, uint8_t *args, uint16_t args_len )
{
    spi_master_transfer_data_t block;

    memset( &block, 0x00, sizeof( spi_master_transfer_data_t ) );

    spi_master_start( &ctx->spi );
    digital_out_low( &ctx->dc );

    block.tx_buffer = &cmd;
    block.tx_length = 1;
    spi_master_transfer( &ctx->spi, &block );

    digital_out_high( &ctx->dc );

    block.tx_buffer = args;
    block.tx_length = args_len;
    spi_master_transfer( &ctx->spi, &block );

    spi_master_stop( &ctx->spi );
}

void
oledc_cfg_setup( oledc_cfg_t *cfg )
{
    cfg->rst = HAL_PIN_NC;
    cfg->dc  = HAL_PIN_NC;
    cfg->en  = HAL_PIN_NC;
    cfg->cs  = HAL_PIN_NC;
    cfg->sck = HAL_PIN_NC;
    cfg->sdo = HAL_PIN_NC;
    cfg->sdi = HAL_PIN_NC;
}

int8_t
oledc_init( oledc_t *ctx, oledc_cfg_t *cfg )
{
    spi_master_config_t spi_cfg;

    memset( &spi_cfg, 0x00, sizeof( spi_master_config_t ) );
    spi_cfg.mode        = SPI_MODE_0;
    spi_cfg.speed       = SPI_MEDIUM;
    spi_cfg.pins.sck    = cfg->sck;
    spi_cfg.pins.mosi   = cfg->sdi;
    spi_cfg.pins.miso   = cfg->sdo;
    spi_cfg.chip_select = cfg->cs;

    if (spi_master_open( &ctx->spi, &spi_cfg ) != SPI_SUCCESS)
    {
        return -1;
    }

    digital_out_init( &ctx->dc, cfg->dc );
    digital_out_high( &ctx->dc );
    digital_out_init( &ctx->rst, cfg->rst );
    digital_out_high( &ctx->rst );
    digital_out_init( &ctx->en, cfg->en );
    digital_out_high( &ctx->en );

    return 0;
}

void
oledc_default_cfg( oledc_t *ctx )
{
    uint8_t arg_value;

    oledc_enable( ctx, 1 );
    oledc_reset( ctx );

    //  Unlock display and turn off.
    arg_value = OLEDC_DEFAULT_OLED_LOCK;
    oledc_command( ctx, OLEDC_COMMAND_LOCK, &arg_value, 1 );
    arg_value = OLEDC_DEFAULT_CMD_LOCK;
    oledc_command( ctx, OLEDC_COMMAND_LOCK, &arg_value, 1 );
    oledc_command( ctx, OLEDC_SLEEP_ON,     0,          0 );

    //  Setup SSD1351.
    arg_value = OLEDC_DEFAULT_REMAP;
    oledc_command( ctx, OLEDC_SET_REMAP,       &arg_value,             1 );
    arg_value = OLEDC_DEFAULT_MUX_RATIO;
    oledc_command( ctx, OLEDC_MUX_RATIO,       &arg_value,             1 );
    arg_value = OLEDC_DEFAULT_START_LINE;
    oledc_command( ctx, OLEDC_SET_START_LINE,  &arg_value,             1 );
    arg_value = OLEDC_DEFAULT_OFFSET;
    oledc_command( ctx, OLEDC_SET_OFFSET,      &arg_value,             1 );
    arg_value = OLEDC_DEFAULT_VCOMH;
    oledc_command( ctx, OLEDC_VCOMH,           &arg_value,             1 );
    arg_value = OLEDC_DEFAULT_DIVSET;
    oledc_command( ctx, OLEDC_CLOCK_DIV,       &arg_value,             1 );
    arg_value = OLEDC_DEFAULT_PRECHARGE;
    oledc_command( ctx, OLEDC_SET_RESET_PRECH, &arg_value,             1 );
    arg_value = OLEDC_DEFAULT_PRECHARGE_2;
    oledc_command( ctx, OLEDC_SETSEC_PRECH,    &arg_value,             1 );
    arg_value = OLEDC_DEFAULT_MASTER_CONT;
    oledc_command( ctx, OLEDC_MASTER_CONTRAST, &arg_value,             1 );
    oledc_command( ctx, OLEDC_CONTRAST,        OLEDC_DEFAULT_CONTRAST, 3 );
    oledc_command( ctx, OLEDC_VSL,             OLEDC_DEFAULT_VSL,      3 );

    //  Set normal mode and turn on display.
    oledc_command( ctx, OLEDC_MODE_NORMAL, 0, 0 );
    oledc_command( ctx, OLEDC_SLEEP_OFF,   0, 0 );
}

void
oledc_fill_screen( oledc_t *ctx, uint16_t color )
{
    box_area( ctx, 0, 0, 96, 96, color );
}

void
oledc_rectangle( oledc_t *ctx, uint8_t col_off, uint8_t row_off,
                 uint8_t col_end, uint8_t row_end, uint16_t color )
{
    box_area( ctx, col_off, row_off, col_end, row_end, color );
}

void
oledc_image( oledc_t *ctx, const uint8_t *img, uint8_t col_off,
             uint8_t row_off )
{
    const uint8_t * ptr = img;
    draw_area( ctx, col_off, row_off, col_off + ptr[ 2 ], row_off + ptr[ 4 ],
               ptr );
}

void
oledc_text( oledc_t *ctx, uint8_t *text, uint16_t x, uint16_t y )
{
    uint8_t * ptr = text;

    if ( ( x >= OLEDC_SCREEN_WIDTH ) || ( y >= OLEDC_SCREEN_HEIGHT ) )
        return;

    ctx->x_coord = x;
    ctx->y_coord = y;

    while (*ptr)
        character( ctx, *ptr++ );
}

void
oledc_set_font( oledc_t *ctx, const uint8_t *font, uint16_t color,
                uint8_t orientation )
{
    ctx->font             = font;
    ctx->font_first_char  = font[ 2 ] + ( font[ 3 ] << 8 );
    ctx->font_last_char   = font[ 4 ] + ( font[ 5 ] << 8 );
    ctx->font_height      = font[ 6 ];
    ctx->font_color       = color;
    ctx->font_orientation = orientation ;
}

void
oledc_fill_area( oledc_t *ctx, uint16_t top, uint16_t left, uint16_t bottom,
                 uint16_t right, uint16_t color )
{
    spi_master_transfer_data_t block;
    uint8_t cmd = OLEDC_WRITE_RAM;
    uint16_t cnt = ( right - left ) * ( bottom - top );
    uint8_t clr[ 2 ] = { 0 };

    oledc_cols[ 0 ] = OLEDC_COL_OFF + left;
    oledc_cols[ 1 ] = OLEDC_COL_OFF + right - 1;
    oledc_rows[ 0 ] = OLEDC_ROW_OFF + top;
    oledc_rows[ 1 ] = OLEDC_ROW_OFF + bottom - 1;
    clr[ 0 ] = color >> 8;
    clr[ 1 ] = color & 0x00FF;

    oledc_command( ctx, OLEDC_SET_COL_ADDRESS, oledc_cols, 2 );
    oledc_command( ctx, OLEDC_SET_ROW_ADDRESS, oledc_rows, 2 );

    memset( &block, 0x00, sizeof( spi_master_transfer_data_t ) );

    spi_master_start( &ctx->spi );
    digital_out_low( &ctx->dc );

    block.tx_buffer = &cmd;
    block.tx_length = 1;
    spi_master_transfer( &ctx->spi, &block );

    digital_out_high( &ctx->dc );

    block.tx_buffer = clr;
    block.tx_length = 2;

    while (cnt--)
    {
        spi_master_transfer( &ctx->spi, &block );
    }

    spi_master_stop( &ctx->spi );
}

void
oledc_begin_frame( oledc_t *ctx, uint16_t top, uint16_t left, uint16_t bottom,
                   uint16_t right )
{
    spi_master_transfer_data_t block;
    uint8_t cmd = OLEDC_WRITE_RAM;

    oledc_cols[ 0 ] = OLEDC_COL_OFF + left;
    oledc_cols[ 1 ] = OLEDC_COL_OFF + right - 1;
    oledc_rows[ 0 ] = OLEDC_ROW_OFF + top;
    oledc_rows[ 1 ] = OLEDC_ROW_OFF + bottom - 1;

    oledc_command( ctx, OLEDC_SET_COL_ADDRESS, oledc_cols, 2 );
    oledc_command( ctx, OLEDC_SET_ROW_ADDRESS, oledc_rows, 2 );

    spi_master_start( &ctx->spi );
    digital_out_low( &ctx->dc );

    memset( &block, 0x00, sizeof( spi_master_transfer_data_t ) );
    block.tx_buffer = &cmd;
    block.tx_length = 1;
    spi_master_transfer( &ctx->spi, &block );
    digital_out_high( &ctx->dc );
}

void
oledc_frame_data( oledc_t *ctx, uint16_t color )
{
    spi_master_transfer_data_t block;
    uint8_t clr[ 2 ] = { 0 };

    memset( &block, 0x00, sizeof( spi_master_transfer_data_t ) );

    clr[ 0 ] = color >> 8;
    clr[ 1 ] = color & 0x00FF;
    block.tx_buffer = clr;
    block.tx_length = 2;
    spi_master_transfer( &ctx->spi, &block );
}

void
oledc_end_frame( oledc_t *ctx )
{
    spi_master_stop( &ctx->spi );
}

// ----------------------------------------------- PRIVATE FUNCTION DEFINITIONS

static void
pixel( oledc_t *ctx, uint8_t col, uint8_t row, uint16_t color )
{
    uint8_t clr[ 2 ] = { 0 };

    if ( ( col > OLEDC_SCREEN_WIDTH ) || ( row > OLEDC_SCREEN_HEIGHT ) )
        return;

    oledc_cols[ 0 ] = OLEDC_COL_OFF + col;
    oledc_cols[ 1 ] = OLEDC_COL_OFF + col;
    oledc_rows[ 0 ] = OLEDC_ROW_OFF + row;
    oledc_rows[ 1 ] = OLEDC_ROW_OFF + row;
    clr[ 0 ] = color >> 8;
    clr[ 1 ] = color & 0x00FF;

    oledc_command( ctx, OLEDC_SET_COL_ADDRESS, oledc_cols, 2 );
    oledc_command( ctx, OLEDC_SET_ROW_ADDRESS, oledc_rows, 2 );
    oledc_command( ctx, OLEDC_WRITE_RAM, clr, 2 );
}

static void
character( oledc_t *ctx, uint16_t ch )
{
    uint8_t ch_width = 0;
    uint8_t x_cnt;
    uint8_t y_cnt;
    uint16_t x = 0;
    uint16_t y = 0;
    uint16_t tmp;
    uint8_t temp = 0;
    uint8_t mask = 0;
    uint32_t offset;
    const uint8_t *ch_table;
    const uint8_t *ch_bitmap;

    if (ch < ctx->font_first_char)
        return;

    if (ch > ctx->font_last_char)
        return;

    offset = 0;
    tmp = ( ch - ctx->font_first_char ) << 2;
    ch_table = ctx->font + 8 + tmp;
    ch_width = *ch_table;

    offset = (uint32_t)ch_table[ 1 ] + ( (uint32_t)ch_table[ 2 ] << 8 ) +
             ( (uint32_t)ch_table[ 3 ] << 16 );

    ch_bitmap = ctx->font + offset;

    if ( ( ctx->font_orientation == OLEDC_FO_HORIZONTAL ) ||
         ( ctx->font_orientation == OLEDC_FO_VERTICAL_COLUMN ) )
    {
        y = ctx->y_coord;
        for (y_cnt = 0; y_cnt < ctx->font_height; y_cnt++)
        {
            x = ctx->x_coord;
            mask = 0;
            for (x_cnt = 0; x_cnt < ch_width; x_cnt++)
            {
                if (!mask)
                {
                    temp = *ch_bitmap++;
                    mask = 0x01;
                }

                if (temp & mask)
                    pixel( ctx, x, y, ctx->font_color );

                x++;
                mask <<= 1;
            }
            y++;
        }

        if (ctx->font_orientation == OLEDC_FO_HORIZONTAL)
            ctx->x_coord = x + 1;
        else
            ctx->y_coord = y;
    }
    else
    {
        y = ctx->x_coord;

        for (y_cnt = 0; y_cnt < ctx->font_height; y_cnt++)
        {
            x = ctx->y_coord;
            mask = 0;

            for (x_cnt = 0; x_cnt < ch_width; x_cnt++)
            {
                if (mask == 0)
                {
                    temp = *ch_bitmap++;
                    mask = 0x01;
                }

                if (temp & mask)
                {
                    pixel( ctx, y, x, ctx->font_color );
                }

                x--;
                mask <<= 1;
            }
            y++;
        }
        ctx->y_coord = x - 1;
    }
}

static void
box_area( oledc_t *ctx, uint8_t start_col, uint8_t start_row, uint8_t end_col,
          uint8_t end_row, uint16_t color )
{
    spi_master_transfer_data_t block;
    uint8_t cmd = OLEDC_WRITE_RAM;
    uint16_t cnt = ( end_col - start_col ) * ( end_row - start_row );
    uint8_t clr[ 2 ] = { 0 };

    if ( ( start_col > OLEDC_SCREEN_WIDTH ) ||
         ( end_col > OLEDC_SCREEN_WIDTH ) )
        return;

    if ( ( start_row > OLEDC_SCREEN_HEIGHT ) ||
         ( end_row > OLEDC_SCREEN_HEIGHT ) )
        return;

    if ( ( end_col < start_col ) ||
         ( end_row < start_row ) )
        return;

    oledc_cols[ 0 ] = OLEDC_COL_OFF + start_col;
    oledc_cols[ 1 ] = OLEDC_COL_OFF + end_col - 1;
    oledc_rows[ 0 ] = OLEDC_ROW_OFF + start_row;
    oledc_rows[ 1 ] = OLEDC_ROW_OFF + end_row - 1;
    clr[ 0 ] = color >> 8;
    clr[ 1 ] = color & 0x00FF;

    oledc_command( ctx, OLEDC_SET_COL_ADDRESS, oledc_cols, 2 );
    oledc_command( ctx, OLEDC_SET_ROW_ADDRESS, oledc_rows, 2 );

    memset( &block, 0x00, sizeof( spi_master_transfer_data_t ) );

    spi_master_start( &ctx->spi );
    digital_out_low( &ctx->dc );

    block.tx_buffer = &cmd;
    block.tx_length = 1;
    spi_master_transfer( &ctx->spi, &block );

    digital_out_high( &ctx->dc );

    block.tx_buffer = clr;
    block.tx_length = 2;

    while (cnt--)
    {
        spi_master_transfer( &ctx->spi, &block );
    }

    spi_master_stop( &ctx->spi );
}

static void
draw_area( oledc_t *ctx, uint8_t start_col, uint8_t start_row, uint8_t end_col,
           uint8_t end_row, const uint8_t *img )
{
    spi_master_transfer_data_t block;
    uint8_t color_buf[ 2 ];
    uint16_t tmp = 0;
    uint8_t cmd = OLEDC_WRITE_RAM;
    uint8_t frb = 0;
    uint8_t srb = 0;
    uint16_t cnt = ( end_col - start_col ) * ( end_row - start_row );

    const uint8_t * ptr = img + OLEDC_IMG_HEAD;

    if ( ( start_col > OLEDC_SCREEN_WIDTH ) ||
         ( end_col > OLEDC_SCREEN_WIDTH ) )
        return;

    if ( ( start_row > OLEDC_SCREEN_HEIGHT ) ||
         ( end_row > OLEDC_SCREEN_HEIGHT ) )
        return;

    if ( ( end_col < start_col ) ||
         ( end_row < start_row ) )
        return;

    oledc_cols[ 0 ] = OLEDC_COL_OFF + start_col;
    oledc_cols[ 1 ] = OLEDC_COL_OFF + end_col - 1;
    oledc_rows[ 0 ] = OLEDC_ROW_OFF + start_row;
    oledc_rows[ 1 ] = OLEDC_ROW_OFF + end_row - 1;

    oledc_command( ctx, OLEDC_SET_COL_ADDRESS, oledc_cols, 2 );
    oledc_command( ctx, OLEDC_SET_ROW_ADDRESS, oledc_rows, 2 );

    memset( &block, 0x00, sizeof( spi_master_transfer_data_t ) );

    spi_master_start( &ctx->spi );
    digital_out_low( &ctx->dc );

    block.tx_buffer = &cmd;
    block.tx_length = 1;
    spi_master_transfer( &ctx->spi, &block );

    digital_out_high( &ctx->dc );

    block.tx_buffer = color_buf;
    block.tx_length = 2;

    while (cnt--)
    {
        color_buf[ 0 ] = ptr[ tmp + 1 ];
        color_buf[ 1 ] = ptr[ tmp ];
        spi_master_transfer( &ctx->spi, &block );
        tmp += 2;
    }

    spi_master_stop( &ctx->spi );
}

// ------------------------------------------------------------------------ END
